@@ -4,11 +4,10 @@ fetch real-time values using Finary API, set targets, and simulate your
 portfolio evolution with optional life events and portfolio operations.
 
 This module is maintained by MadeInPierre.
-You can always get the latest version of this module at:
-    https://github.com/madeinpierre/finalynx
+You can always get the latest version of this module on
+[GitHub](https://github.com/madeinpierre/finalynx).
 """
 # flake8: noqa
-# Import meta information such as version, author, copyright, etc.
 import os
 import sys
 
@@ -16,7 +15,7 @@ from .__meta__ import __author__  # noqa: F401
 from .__meta__ import __copyright__  # noqa: F401
 from .__meta__ import __version__  # noqa: F401
 
-# add finary_api submodule to path to allow imports to work
+# Add finary_api submodule to path to allow imports to work
 sys.path.append(os.path.join(os.path.dirname(__file__), "finary_api"))
 
 # Portfolio
@@ -25,11 +24,14 @@ from .portfolio import Line, Folder, Bucket, SharedFolder, Portfolio, FolderDisp
 from .portfolio import console
 
 # Fetch
-from .fetch import finary_fetch
+from .fetch import FetchFinary
 
 # Advisor
 from .copilot import Copilot
 
+# Dashboard
+from .dashboard import Dashboard
+
 # Simulator
 from .simulator import Simulator
 
