@@ -1,7 +1,10 @@
 import datetime
 import json
 import os
+from typing import Any
+from typing import Callable
 from typing import Dict
+from typing import List
 from typing import Optional
 from typing import Tuple
 
@@ -10,8 +13,9 @@ from rich.prompt import Confirm
 from rich.tree import Tree
 from unidecode import unidecode
 
-import finary_api.__main__ as ff  # type: ignore
-import finary_api.constants  # type: ignore
+import finary_api.__main__ as ff
+import finary_api.constants
+import finary_api.user_portfolio
 from ..console import console
 from ..portfolio.line import Line
 from ..portfolio.portfolio import Portfolio
@@ -23,6 +27,20 @@ class FetchFinary(Fetch):  # TODO update docstrings
 
     CACHE_FILENAME = "finary_data.json"
 
+    _categories = [
+        "fiats",
+        "securities",
+        "cryptos",
+        "fonds_euro",
+        "startups",
+        "precious_metals",
+        "scpis",
+        "generic_assets",
+        "real_estates",
+        "loans",
+        "crowdlendings",
+    ]
+
     def __init__(
         self, portfolio: Portfolio, clear_cache: bool = False, force_signin: bool = False, ignore_orphans: bool = False
     ) -> None:
@@ -72,13 +90,13 @@ class FetchFinary(Fetch):  # TODO update docstrings
             os.remove(self.cache_fullpath)
 
         # This will hold a key:amount dictionary of all lines found in the Finary account
-        lines_dict: Dict[str, int] = self._get_cache()  # try to get the data in the cache first
+        lines_list: List[Dict[str, Any]] = self._get_cache()  # try to get the data in the cache first
         tree = Tree("Finary API", highlight=True, hide_root=True)
 
         # If the cache is not empty, Match all lines to the portfolio hierarchy
-        if lines_dict:
-            for key, value in lines_dict.items():
-                self._match_line(lines_dict, tree, key, value)
+        if lines_list:
+            for line in lines_list:
+                self._match_line([], tree, key=line["key"], id=line["id"], amount=line["amount"])
 
         # If there's no valid cache, signin and fetch the data online
         else:
@@ -87,13 +105,13 @@ class FetchFinary(Fetch):  # TODO update docstrings
                 return Tree("Finary signin failed.")
 
             try:
-                lines_dict, tree = self._fetch_data(session, tree)
+                lines_list, tree = self._fetch_data(session, tree)
             except Exception:
                 console.log("[red bold]Error: Couldn't fetch data, please try using the `-f` option to signin again.")
                 return tree
 
             # Save what has been found in a cache file for offline use and better performance at next launch
-            self._save_cache(lines_dict)
+            self._save_cache(lines_list)
 
         # Return a rich tree to be displayed in the console as a recap of what has been fetched
         console.log("Done fetching Finary data.")
@@ -177,85 +195,109 @@ class FetchFinary(Fetch):  # TODO update docstrings
 
             return session
 
-    def _fetch_data(self, session: Session, tree: Tree) -> Tuple[Dict[str, int], Tree]:
+    def _fetch_data(self, session: Session, tree: Tree) -> Tuple[List[Dict[str, Any]], Tree]:
         """Internal method used to fetch every investment in your Finary account.
         :returns: A dictionary of all fetched investments (name:amount format), and a `Tree`
         instance which can be displayed in the console to make sure everything was retrieved.
         """
         # Create a rich Tree to display the fetched data nicely
-        lines_dict: Dict[str, int] = {}
+        lines_list: List[Dict[str, Any]] = []
 
-        # Comptes courants, Livrets et Fonds euro
-        checkings = ff.get_checking_accounts(session, "1w")["result"]
-        savings = ff.get_savings_accounts(session, "1w")["result"]
-        fonds = ff.get_fonds_euro(session, "1w")["result"]
-        for result, name in zip([checkings, savings, fonds], ["Comptes courants", "Livrets", "Fonds euro"]):
-            console.log(f"Fetching {name.lower()}...")
-            node = tree.add("[bold]" + str(round(result["timeseries"][-1][1])) + " " + name)
-            for k, e in result["distribution"].items():
-                self._match_line(lines_dict, node, k, e["amount"])
+        # Common fetching procedure for multiple sources
+        def _process(
+            step_name: str,
+            function: Callable[..., Dict[str, Any]],
+            callback: Callable[[Dict[str, Any]], Tuple[str, str, int]],
+            period: bool = True,
+        ) -> None:
+            console.log(f"Fetching {step_name.lower()}...")
+            node = tree.add(f"[bold]{step_name}")
+            result = function(session, "1w")["result"]["data"] if period else function(session)["result"]
 
-        # Autres
-        console.log("Fetching other assets...")
-        other = ff.get_other_assets(session, "1w")["result"]
-        f_other_total = round(other["timeseries"][-1][1])
-        node = tree.add("[bold]" + str(round(f_other_total)) + " Autres")
-        for item in other["data"]:
-            self._match_line(lines_dict, node, item["name"], item["current_value"])
+            for e in result:
+                key, id, amount = callback(e)
+                self._match_line(lines_list, node, key, id, round(amount))
 
-        # Investissements
+        # Process similar sources together
+        _process("Checkings", ff.get_checking_accounts, lambda e: (e["name"], e["id"], e["fiats"][0]["current_value"]))
+        _process("Savings", ff.get_savings_accounts, lambda e: (e["name"], e["id"], e["fiats"][0]["current_value"]))
+        _process("Fonds euro", ff.get_fonds_euro, lambda e: (e["name"], e["id"], e["current_value"]))
+        _process("Others", ff.get_other_assets, lambda e: (e["name"], e["id"], e["current_value"]))
+        _process(
+            "Precious metals",
+            ff.get_user_precious_metals,
+            lambda e: (e["precious_metal"]["name"], e["id"], e["current_value"]),
+            period=False,
+        )
+
+        # Investments
         console.log("Fetching investments...")
+        node = tree.add("[bold]Investments")
         investments = ff.get_portfolio_investments(session)["result"]
-        f_invest_total = round(investments["total"]["amount"])
-        node = tree.add("[bold]" + str(round(f_invest_total)) + " Investissements")
         for account in investments["accounts"]:
             node_account = node.add("[bold]Account: " + account["name"])
-            for category in [
-                "fiats",
-                "securities",
-                "cryptos",
-                "fonds_euro",
-                "startups",
-                "precious_metals",
-                "generic_assets",
-                "loans",
-                "crowdlendings",
-            ]:
-                for item in account[category]:
-                    self._match_line(lines_dict, node_account, item["security"]["name"], item["current_value"])
+            for account in account["securities"]:
+                self._match_line(
+                    lines_list,
+                    node_account,
+                    key=account["security"]["name"],
+                    id=account["id"],
+                    amount=account["current_value"],
+                )
 
-        # Immobilier
+        # Cryptos
+        console.log("Fetching cryptos...")
+        node = tree.add("[bold]Cryptos")
+        cryptos = finary_api.user_portfolio.get_portfolio_cryptos(session)["result"]
+
+        for account in cryptos["accounts"]:
+            node_account = node.add("[bold]Account: " + account["name"])
+            for account in account["cryptos"]:
+                self._match_line(
+                    lines_list,
+                    node_account,
+                    key=account["crypto"]["name"],
+                    id=account["id"],
+                    amount=account["current_value"],
+                )
+
+        # Real estate
         console.log("Fetching real estate...")
+        node = tree.add("[bold]Real estate")
         real_estate = ff.get_real_estates(session, "1w")["result"]
-        f_re_total = round(real_estate["total"]["amount"])
-        node = tree.add("[bold]" + str(round(f_re_total)) + " Immobilier")
 
         for item in real_estate["data"]["real_estates"]:
-            self._match_line(lines_dict, node, item["description"], item["current_value"])
-
+            self._match_line(lines_list, node, key=item["description"], id=item["id"], amount=item["current_value"])
         for item in real_estate["data"]["scpis"]:
-            self._match_line(lines_dict, node, item["scpi"]["name"], item["current_value"])
+            self._match_line(
+                lines_list, node, key=item["scpi"]["name"], id=item["scpi"]["id"], amount=item["current_value"]
+            )
 
-        return lines_dict, tree
+        return lines_list, tree
 
-    def _match_line(self, lines_dict: Dict[str, int], node: Tree, key: str, amount: int) -> None:
+    def _match_line(self, lines_list: List[Dict[str, Any]], node: Tree, key: str, id: str, amount: int) -> None:
         """Internal method used to register a new investment found from Finary."""
 
         # Discard non-ASCII characters in the key
-        key, amount = unidecode(key), round(amount)
+        key, id, amount = unidecode(key), str(id), round(amount)
 
         # Add the line to the dictionary of fetched items
-        lines_dict[key] = amount
+        lines_list.append({"key": key, "id": id, "amount": amount})
 
         # Add the line to the rendering tree
-        node_child = node.add(f"{amount} {key}")
+        node_child = node.add(f"{amount} {key} [dim white]{id=}")
 
-        # Fill the line's amount in the portfolio, attach it to root if the line is absent (unless ignored)
-        if not self.portfolio.set_child_amount(key, amount) and not self.ignore_orphans:
+        # Fill the line's amount in the portfolio
+        for k in [key, id]:
+            if self.portfolio.set_child_amount(k, amount):
+                return
+
+        # Attach it to root if the line is absent (unless ignored)
+        if not self.ignore_orphans:
             node_child.add("[yellow]WARNING: This line did not match with any envelope, attaching to root")
             self.portfolio.add_child(Line(key, amount=amount))
 
-    def _get_cache(self) -> Dict[str, int]:
+    def _get_cache(self) -> List[Dict[str, Any]]:
         """Attempt to retrieve the cached data. Check if more than an hour has passed since the last update.
         :returns: A key:amount dictionary if the cache file is less than an hour old, None otherwise.
         """
@@ -263,7 +305,7 @@ class FetchFinary(Fetch):  # TODO update docstrings
         # Abort retrieving cache if the file doesn't exist
         if not os.path.exists(self.cache_fullpath):
             console.log("No cache file found, fetching data.")
-            return {}
+            return []
 
         # Parse the JSON content
         with open(self.cache_fullpath) as f:
@@ -276,12 +318,12 @@ class FetchFinary(Fetch):  # TODO update docstrings
 
         if hours_passed < self.MAX_CACHE_HOURS:
             console.log(f"Using recently cached data (<{self.MAX_CACHE_HOURS}h max)")
-            lines: Dict[str, int] = data["lines"]
+            lines: List[Dict[str, Any]] = data["lines"]
             return lines
         console.log(f"Fetching data (cache file is {hours_passed}h old > {self.MAX_CACHE_HOURS}h max)")
-        return {}
+        return []
 
-    def _save_cache(self, lines_dict: Dict[str, int]) -> None:
+    def _save_cache(self, lines_list: List[Dict[str, Any]]) -> None:
         """Save the fetched data locally to work offline and reduce the amoutn of calls to the API.
         :param tree: Generated tree object containing all information
         """
@@ -289,6 +331,6 @@ class FetchFinary(Fetch):  # TODO update docstrings
         # Save current date and time to a JSON file with the fetched data
         console.log(f"Saving fetched data in '{self.cache_fullpath}'")
         current_time = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
-        data = {"last_updated": current_time, "lines": lines_dict}
+        data = {"last_updated": current_time, "lines": lines_list}
         with open(self.cache_fullpath, "w") as f:
             json.dump(data, f, indent=4)
